# Copyright (c) Abel Deuring 2013
#
# tftools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.


"""Create instances of brick/bricklet classes without passing a UID.

Brick/bricklet constructors require two parameters: A UID string
and an IPConnection.

Passing the UID is obviously necessary to distiguish between two or
more devices of the same type, but it is unnecessaryly cumbersome
to look up the UIDs if only one device of a given type is used.

This class allows the creation brick/bricklet class instances
by simply passing a class to a factory method.


Using this module
=================

    >>> from brick_factory import BrickFactory
    >>> from tinkerforge.brick_master import BrickMaster
    >>> from tinkerforge.ip_connection import IPConnection

    >>> DEFAULT_HOST = 'localhost'
    >>> DEFAULT_PORT = 4223

    >>> conn = IPConnection()
    >>> conn.connect(DEFAULT_HOST, DEFAULT_PORT)
    >>> factory = BrickFactory([conn])
    >>> try:
    >>>     factory.find_devices(conn)
    >>>     master, uid = factory.make_brick(BrickMaster)
    >>>     # ...
    >>> finally:
    >>>     conn.disconnect()

You can pass more than one connection if needed:

    >>> conn1 = IPConnection()
    >>> conn1.connect(DEFAULT_HOST, DEFAULT_PORT)
    >>> conn2 = IPConnection()
    >>> conn2.connect('192.168.1.42', 4223)
    >>> factory = BrickFactory([conn1, conn2])
"""

from functools import partial
from importlib import import_module
from new import classobj
import os.path
from pkgutil import iter_modules
from time import sleep
import tinkerforge
from tinkerforge.ip_connection import Device, IPConnection


__all__ = ['BrickFactory', 'BrickFactoryError']


class BrickFactoryError(Exception):
    pass


class BrickFactory:
    """Create instances of brick/bricklet classes without passing a UID."""
    def __init__(self, connections, wait=0.5):
        self.device_map = {}
        self.find_devices(connections, wait)

    def find_devices(self, connections, wait=0.5):
        for conn in connections:
            callback = partial(self._enumerated, conn)
            conn.register_callback(IPConnection.CALLBACK_ENUMERATE, callback)
            conn.enumerate()
        sleep(wait)
        for conn in connections:
            del conn.registered_callbacks[IPConnection.CALLBACK_ENUMERATE]

    def _enumerated(self, conn, uid, connected_uid, position, hw_version,
                  fw_version, dev_id, enum_type):
        self.device_map.setdefault(dev_id, {})[uid] = (
            connected_uid, position, hw_version, fw_version, enum_type, conn,
            dev_id)

    def make_brick(self, cls, *args):
        known_devices = self.device_map.get(cls.DEVICE_IDENTIFIER)
        if not known_devices:
            raise BrickFactoryError(
                "No devices found for class %s" % cls.__name__)
        if len(known_devices) > 1:
            dev_list = ', '.join(d[0] for d in known_devices)
            raise BrickFactoryError(
                "More than one device for class %s found." % cls.__name__)
        uid, info = known_devices.items()[0]
        return cls(uid, info[5], *args), uid

    def device_class_uid_match(self, cls, uid):
        """Check if the given class is suitable for the given UID.

        Not needed for the main purpose of BrickFactory, but a useful check
        for other use cases to avoid to create device class instance
        where the class does not match the type of the given device.
        """
        for _, info in self.device_map.items():
            if uid in info:
                return info[uid][6] == cls.DEVICE_IDENTIFIER

        raise BrickFactoryError('Device %s not found.' % uid)


class BrickClassRegistry:
    def __init__(self, packages=[tinkerforge]):
        self.class_map = {}
        self.type_map = {}
        for package in packages:
            prefix = '%s.' % package.__name__
            for importer, name, is_pkg in iter_modules(
                    [os.path.dirname(package.__file__)], prefix):
                module = import_module(name)
                for item in module.__dict__.values():
                    if (
                            type(item) is classobj and issubclass(item, Device)
                            and item is not Device):
                        self.class_map[item.__name__] = item
                        # Note that many or all device classes are available
                        # under two names, like 'BrickDC' and 'DC', but we
                        # want only one entry for each real class.
                        type_entry = self.type_map.setdefault(
                            item.DEVICE_IDENTIFIER, [])
                        if item not in type_entry:
                            type_entry.append(item)

    def __call__(self, class_name):
        if class_name not in self.class_map:
            raise ValueError('class %s is not registered' % class_name)
        return self.class_map[class_name]

    def cls_for_type(self, type_):
        if type_ in self.type_map:
            return self.type_map[type_]
        else:
            raise BrickFactoryError('No class registered for type %s' % type_)
