#!/usr/bin/python

# Copyright Abel Deuring 2013
#
# tftools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.


__metaclass__ = type


import argparse
import csv
from datetime import datetime
from pyparsing import (
    alphas, alphanums, Combine, delimitedList, nums, Optional, restOfLine,
    Word,
    )
from time import sleep, time
import tftools
import tinkerforge
from tinkerforge.ip_connection import IPConnection
import os
import sys
from tftools.brick_factory import BrickClassRegistry, BrickFactory


DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 4223


class LoggerError(Exception):
    pass


def setup_connection(args):
    if args.host is not None:
        host = args.host
    else:
        host = DEFAULT_HOST
    if args.port is not None:
        port = args.port
    else:
        port = DEFAULT_PORT
    conn = IPConnection()
    conn.connect(host, port)
    return conn


class LoggerDevice:

    def __init__(self, conn, factory, registry, info):
        self.readers = []
        self.writers = []
        self.get_params = info['get']
        self.set_params = info['set']

        class_name = info['class_name']
        try:
            tf_class = registry(class_name)
        except ValueError, e:
            expected = 'class %s is not registered' % class_name
            if str(e) == expected:
                raise LoggerError(expected)
            else:
                raise
        if info['args'] and info['args'][0] != '_':
            uid = info['args'][0]
            if not factory.device_class_uid_match(tf_class, uid):
                raise LoggerError(
                    'Class %s is not usable for device %s' % (class_name, uid))
            self.tf_device = tf_class(uid, conn, *info['args'][1:])
            self.uid = uid
        else:
            self.tf_device, self.uid = factory.make_brick(
                tf_class, *info['args'][1:])
        for name in self.get_params:
            method = getattr(self.tf_device, 'get_%s' % name, None)
            if method is None:
                raise LoggerError(
                    'cannot read parameter %s for class %s'
                    % (name, class_name))
            self.readers.append(method)
        for name, value in self.set_params:
            method = getattr(self.tf_device, 'set_%s' % name, None)
            if method is None:
                raise LoggerError(
                    'cannot write parameter %s for class %s'
                    % (name, class_name))
            self.writers.append((method, value))

    def call_writers(self):
        [m(v) for m, v in self.writers]

    def read_values(self):
        return [r() for r in self.readers]

    def get_config(self):
        """Return the config string for this device."""
        set_params = [
            '%s=%s' % (name, value) for name, value in self.set_params]
        all_params = self.get_params + set_params
        all_params = ','.join(all_params)
        return '%s(%s):%s' % (
            self.tf_device.__class__.__name__, self.uid, all_params)


def init_devices(args, device_info, conn, factory, registry):
    return [
        LoggerDevice(conn, factory, registry, info) for info in device_info]


def call_device_writers(devices):
    for d in devices:
        d.call_writers()


def read_values(devices):
    result = [d.read_values() for d in devices]
    return sum(result, [])


def get_configs(devices):
    return [d.get_config() for d in devices]


class Output:

    def __init__(self, output_spec):
        pass

    @classmethod
    def known_output_spec(cls, output_spec):
        """Return True if the given output_spec is recognized by this
        output class, else return False.
        """
        raise NotImplemented

    def write_info(self, arg_info, device_info):
        """Write metadata about this log run."""
        raise NotImplemented

    def write_data(self, data):
        """Write one data set."""
        raise NotImplemented

    def close(self):
        pass

    def remove(self):
        pass


class CSVOutput(Output):

    def __init__(self, output_spec):
        if not self.known_output_spec(output_spec):
            raise LoggerError(
                'Invalid output spec for class %s: %s'
                % (self.__class__.__name__, output_spec))
        if output_spec.startswith('csv:'):
            output_spec = output_spec[len('csv:'):]
        self._filename = output_spec
        self._file = open(output_spec, 'w')
        self.writer = csv.writer(self._file)

    @classmethod
    def known_output_spec(cls, output_spec):
        if ':' in output_spec:
            return output_spec.startswith('csv:')
        return output_spec.lower().endswith('.csv')

    def close(self):
        self._file.close()

    def remove(self):
        # cleanup call; should only be used before any data is written.
        self.close()
        os.unlink(self._filename)

    def write_info(self, arg_info, device_info):
        for info in arg_info:
            self.writer.writerow(info)
        for info in device_info:
            self.writer.writerow([info])
        self.writer.writerow([])

    def write_data(self, data):
        self.writer.writerow(data)


class TextOutput(Output):

    def __init__(self, output_spec):
        if not self.known_output_spec(output_spec):
            raise LoggerError(
                'Invalid output spec for class %s: %s'
                % (self.__class__.__name__, output_spec))
        if output_spec.startswith('text:'):
            output_spec = output_spec[len('text:'):]
        self._filename = output_spec
        if output_spec in ('-', ''):
            self._file = sys.stdout
        else:
            self._file = open(output_spec, 'w')

    @classmethod
    def known_output_spec(cls, output_spec):
        if ':' in output_spec:
            return output_spec.startswith('text:')
        return output_spec.lower().endswith('.txt')

    def close(self):
        if self._file != sys.stdout:
            self._file.close()

    def remove(self):
        self.close()
        if self._file != sys.stdout:
            os.unlink(self._filename)

    def write_info(self, arg_info, device_info):
        for info in arg_info:
            self._file.write('%s: %s\n' % info)
        for info in device_info:
            self._file.write('%s\n' % info)
        self._file.write('\n')

    def write_data(self, data):
        data[0] = datetime.fromtimestamp(data[0])
        data = [str(d) for d in data]
        data = ' '.join(data)
        self._file.write('%s\n' % data)


output_classes = [CSVOutput, TextOutput]


def get_writer(spec):
    for cls in output_classes:
        if cls.known_output_spec(spec):
            return cls(spec)
    raise LoggerError('No output class found for spec %s' % spec)


def open_output(args, devices):
    output = []
    try:
        for spec in args.output:
            output.append(get_writer(spec))
    except Exception:
        for writer in output:
            writer.remove()
        raise
    param_names = (
        'comment', 'host', 'port', 'repeat_time', 'start_time', 'stop_time')
    arg_info = [
        (name, getattr(args, name))
        for name in param_names
        if name in args and getattr(args, name) is not None
    ]
    device_info = [d.get_config() for d in devices]
    for o in output:
        o.write_info(arg_info, device_info)
    return output


def record(devices, output):
    data = [d.read_values() for d in devices]
    data = sum(data, [time()])
    [o.write_data(data) for o in output]


def acquire(args, devices, output):
    # XXX check needed that repeat_time is a positive number
    if args.start_time is not None and args.start_time > 0:
        sleep(args.start_time - time())
    [d.call_writers() for d in devices]
    delta_t = args.repeat_time
    start = time()
    next_ = start + delta_t
    if args.stop_time is not None and args.stop_time > 0:
        stop = start + args.stop_time
    else:
        stop = None
    while stop is None or stop > time():
        record(devices, output)
        wait = next_ - time()
        if wait > 0:
            sleep(wait)
        else:
            print >>sys.stderr, "sample time missed by %s" % wait
        next_ += delta_t


def parse_device_param(text):
    # text should have one of the following formats
    #    BrickClassName
    #    BrickClassName(uid)
    #    BrickClassName(uid [, arg...])
    #    BrickClassName:read_value1,read_value2,write_param1=23...
    #    BrickClassName(uid [, arg...]):read_value1,write_param1=23...

    def register_params(s, loc, tokens):
        while tokens:
            if len(tokens) > 1:
                if tokens[1] == '=':
                    name = tokens.pop(0)
                    tokens.pop(0)
                    value = tokens.pop(0)
                    try:
                        value = int(value)
                    except ValueError:
                        try:
                            value = float(value)
                        except ValueError:
                            raise LoggerError(
                                'Invalid parameter value: %s'
                                % channel_info[1])

                    collected_set.append((name, value))
                    continue
            collected_get.append(tokens.pop(0))

    args = delimitedList(Word(alphanums + '_.'))('args')
    word = Word(alphas + '_', alphanums + '_')
    class_spec = word('class_name') + Optional('(' + args + ')')
    num = Combine(
        Optional(Word('+-')) + Word(nums) +
        Optional(Combine('.' + Optional(Word(nums)))))
    get_param = word
    set_param = word + '=' + num
    params = delimitedList(get_param ^ set_param)
    collected_get = []
    collected_set = []
    params.setParseAction(register_params)
    parser = class_spec + Optional(':' + params) + restOfLine('remain')
    parsed = parser.parseString(text)

    if parsed['remain']:
        raise LoggerError(
            'Unparsed device spec: %s (full text: %s'
            % (parsed['remain'], text))
    result = {
        'class_name': parsed['class_name'],
        'args': list(parsed.get('args', [])),
        'get': collected_get,
        'set': collected_set,
        }
    return result

    # ret: class_name, uid, list of get params, list of set params
    # ex 'BrickDC', '123', ['val1', 'val2'], [('p1', 12), ('p2', 42)]
    return parts[0], parts[1], get_params, set_params


def get_params(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-r', '--repeat-time', type=int, dest='repeat_time', default=1)
    parser.add_argument(
        '-s', '--start-time', type=int, dest='start_time')
    parser.add_argument(
        '-t', '--stop-time', type=int, dest='stop_time')
    parser.add_argument(
        '-o', '--output', dest='output', nargs='+', required=True)
    parser.add_argument(
        '-H', '--host', dest='host')
    parser.add_argument(
        '-p', '--port', type=int, dest='port')
    parser.add_argument(
        '-c', '--comment', dest='comment')
    parser.add_argument('device', nargs='+')
    args = parser.parse_args(argv)
    device_info = []
    for device in args.device:
        device_info.append(parse_device_param(device))
    return args, device_info


def main(argv=sys.argv):
    args, device_info = get_params(argv[1:])
    conn = setup_connection(args)
    output = []
    try:
        factory = BrickFactory([conn])
        registry = BrickClassRegistry([tinkerforge, tftools])
        devices = init_devices(args, device_info, conn, factory, registry)
        output = open_output(args, devices)
        acquire(args, devices, output)
    except LoggerError, e:
        print >>sys.stderr, e
        return 1
    finally:
        conn.disconnect()
        for o in output:
            o.close()
    return 0


if __name__ == '__main__':
    sys.exit(main())
