# -*- coding: utf-8 -*-
# Variant of BrickletVoltageCurrent that is aware of different shunts.

from ConfigParser import ConfigParser
import os.path
import sys
from tinkerforge.bricklet_voltage_current import BrickletVoltageCurrent

class BrickletVoltageCurrentVariableShunt(BrickletVoltageCurrent):
    """A class for a modified voltage-current bricklet.

    This class assumes that the original shunt of the bricklet has been
    replaced with a set of selectable shunts. This allows to extend
    the measurement range both to higher and lower currents.

    Since the firmware is not futher modified and since each shunt
    requires its own calibration, this class reads the calibration data
    from an ordinary INI file.

    The contents should look like this:

    # Bricklet ID ffb, 20A range, 4mOhm shunt
    [ffb:20A]
    multiplier=20
    divisor=32
    unit=1mA

    # Bricklet ID ffb, 2A range, 40mOhm shunt
    [ffb:2A]
    multiplier=20
    divisor=32
    unit=0.1mA

    Each section specifies the calibration for one bricklet and one shunt.
    The name of the section consists of the bricklet's UID and a name
    for the range, separated by a ':'.

    The values above are correct for shunts that have a resistance of
    exactly 4mOhm and 40mOhm, respectively -- if the internal calibration
    data in the bricklet's EEPROM is set to (0, 0). In this case, a
    call of BrickletVoltageCurrent.get_current() returns the raw ADC
    value.

    It might make more sense to use the same units for all ranges.
    In this case, use these calibration values:

    # Bricklet ID ffb, 20A range, 4mOhm shunt
    [ffb:20A]
    multiplier=200
    divisor=32
    unit=0.1mA

    # Bricklet ID ffb, 2A range, 40mOhm shunt
    [ffb:2A]
    multiplier=20
    divisor=32
    unit=0.1mA
    """

    def __init__(self, uid, ipconn, range_name, calib_file=None):
        BrickletVoltageCurrent.__init__(self, uid, ipconn)
        self._uid_b58 = uid
        self.get_config(calib_file)
        if range_name not in self.ranges:
            all_ranges = ', '.join(sorted(self.ranges))
            raise ValueError(
                'Measurement range %s is not available. Defined ranges: %s'
                % (range_name, all_ranges))
        self.range_name = range_name
        self.multiplier = self.ranges[range_name]['multiplier']
        self.divisor = self.ranges[range_name]['divisor']
        self.unit = self.ranges[range_name]['unit']

    def get_config(self, calib_file):
        if calib_file is None:
            calib_file = os.path.expanduser(
                '~/.tftools/BrickletVoltageCurrentVariableShunt.ini')
        self.config = {}
        config = ConfigParser()
        config.readfp(open(calib_file))
        section_start = self._uid_b58 + ':'
        self.ranges = {}
        for section in config.sections():
            if section.startswith(section_start):
                range_name = section[len(section_start):]
                range_info = {}
                for name in ('multiplier', 'divisor'):
                    range_info[name]=config.getint(section, name)
                range_info['unit'] =  config.get(section, 'unit')
                self.ranges[range_name] = range_info

    def get_current(self):
        result = BrickletVoltageCurrent.get_current(self)
        return result * self.multiplier / self.divisor

    def register_callback(self, id, callback):
        if id == self.CALLBACK_CURRENT:
            self.user_cb_current = callback
            BrickletVoltageCurrent.register_callback(id, self.cb_current)
        else:
            BrickletVoltageCurrent.register_callback(id, callback)

    def cb_current(self, value):
        self.user_cb_current(self.multiplier / self.divisor)

    def _reset_eprom_calibration(self):
        self.set_calibration(0, 0)


if __name__ == '__main__':
    from tinkerforge.ip_connection import IPConnection

    if '-H' in sys.argv:
        host_name = sys.argv[sys.argv.index('-H')+1]
    else:
        host_name = 'localhost'
    if '-p' in sys.argv:
        port = int(sys.argv[sys.argv.index('-p')+1])
    else:
        port = 4223
    conn = IPConnection()
    conn.connect(host_name, port)
    try:
        b = BrickletVoltageCurrentVariableShunt(sys.argv[-1], conn, '20A')
    finally:
        conn.disconnect()
